try:
    from hic3defdr.analysis import HiC3DeFDR
    from hic3defdr.plotting.roc import plot_roc
    from hic3defdr.plotting.fdr import plot_fdr
    from hic3defdr.plotting.fn_vs_fp import plot_fn_vs_fp
    from hic3defdr.plotting.distance_bias import plot_distance_bias
    from hic3defdr.plotting.dispersion import compare_disp_fits

    __all__ = [
        'HiC3DeFDR',
        'plot_roc',
        'plot_fdr',
        'plot_fn_vs_fp',
        'plot_distance_bias',
        'compare_disp_fits'
    ]
except ImportError:
    print('WARNING: hic3defdr has not been installed successfully yet')

from hic3defdr._version import get_version
__version__ = get_version()



import numpy as np
import scipy.sparse as sparse

from hic3defdr.util.clusters import find_clusters


def classify(row, col, value, clusters):
    """
    Classifies pixels (specified by row, col coordinates) into classes
    (corresponding to the columns of value) where their value is highest,
    ignoring those pixels that don't belong to the specified clusters.

    Parameters
    ----------
    row, col : np.ndarray
        Row and column indices, respectively, identifying pixels which
        correspond to rows of ``value``.
    value : np.ndarray
        Value to use to assign classes. Rows correspond to pixels, columns to
        conditions.
    clusters : list of set of tuple
        Only pixels in these clusters will be assigned classes.

    Returns
    -------
    list of list of set of tuple
        The outer list represents classes (one per condition). Its elements
        represent clusters of points with that class/condition.
    """
    # convert clusters to idx, parallel to row/col
    pixels = set().union(*(c for c in clusters))
    idx = np.zeros_like(row, dtype=bool)
    for i, p in enumerate(zip(row, col)):
        if p in pixels:
            idx[i] = True

    # assign classes
    classes = np.argmax(value[idx, :], axis=1)

    # re-cluster each class
    n = max(row.max(), col.max()) + 1
    class_clusters = []
    for c in range(value.shape[1]):
        coo = sparse.coo_matrix(
            (np.ones((classes == c).sum()),
             (row[idx][classes == c], col[idx][classes == c])), shape=(n, n))
        class_clusters.append(find_clusters(coo))

    return class_clusters



import numpy as np
import scipy.sparse as sparse

from lib5c.util.system import check_outdir


def roll_footprint(footprint):
    """
    "Rolls" each row to the right by its row index, then reverses the row order
    and transposes the result. This is equivalent to the operation
    ``f(x) = sparse.dia_matrix(x).data`` where ``x`` is a dense matrix, as long
    as ``sparse.dia_matrix(x).offsets`` is a sequential matrix (this is
    guaranteed by the BandedMatrix subclass).

    The result of this is that the rolled footprint is suitable for use in
    convolution against ``sparse.dia_matrix.data``, for example as
    ``convolve(sparse.dia_matrix(x).data, roll_footprint(footprint))`` where
    ``x`` is a dense matrix and ``footprint`` is expressed in terms of the
    dense spatial coordinates.

    Parameters
    ----------
    footprint : np.ndarray
        The footprint to roll.

    Returns
    -------
    np.ndarray
        The rolled footprint.
    """
    n = len(footprint)
    m = 2*n - 1
    row_idx, col_idx = np.ogrid[:n, :m]
    col_idx = (col_idx + np.arange(n)[:, np.newaxis]) % m
    return np.hstack(
        [np.zeros((n, n-1), dtype=int), footprint])[row_idx, col_idx][::-1, :].T


class BandedMatrix(sparse.dia_matrix):
    def __init__(self, data, shape=None, copy=False, max_range=300,
                 upper=None):
        if shape is None:
            if isinstance(data, BandedMatrix):
                instance = data
            elif isinstance(data, sparse.dia_matrix):
                if not self.is_bandedmatrix(data):
                    raise ValueError('dia_matrix detected, but not a '
                                     'BandedMatrix')
                instance = self.from_dia_matrix(data, copy=copy)
            elif isinstance(data, sparse.spmatrix):
                instance = self.from_sparse(data, max_range=max_range,
                                            upper=upper)
            elif isinstance(data, np.ndarray):
                instance = self.from_ndarray(data, max_range=max_range,
                                             upper=upper)
            else:
                raise ValueError('invalid arguments passed to BandedMatrix '
                                 'constructor')
            self.__init__((instance.data, instance.offsets),
                          shape=instance.shape, copy=False)
        else:
            super(BandedMatrix, self).__init__(
                data, shape=shape, copy=copy)

    @classmethod
    def from_dia_matrix(cls, dia_matrix, copy=True):
        if copy:
            return cls((dia_matrix.data.copy(), dia_matrix.offsets.copy()),
                       shape=dia_matrix.shape)
        else:
            return cls((dia_matrix.data, dia_matrix.offsets),
                       shape=dia_matrix.shape)

    @classmethod
    def from_ndarray(cls, ndarray, max_range=300, upper=None):
        if upper is None:
            upper = np.all(np.tril(ndarray, k=-1) == 0)
        max_range = min(ndarray.shape[0], max_range)
        data = np.pad(
            cls.roll_matrix(ndarray.astype(float).T, max_range=max_range),
            ((1, 1), (0, 0)), mode='constant', constant_values=np.nan)
        offsets = np.arange(-max_range-1, max_range+2)
        instance = cls((data, offsets), shape=ndarray.shape)
        if upper:
            instance.make_upper()
        return instance

    @classmethod
    def from_sparse(cls, spmatrix, max_range=300, upper=None):
        """
        Constructs a BandedMatrix from a scipy.spare.spmatrix instance.

        Parameters
        ----------
        spmatrix : scipy.spare.spmatrix
            The sparse matrix to convert to BandedMatrix.
        max_range : int
            Offdiagnals beyond the max_range'th offdiagonal will be discarded.
        upper : bool, optional
            Pass True to keep only the upper triangular entries. Pass False to
            keep all the entries. Pass None to guess what to do based on whether
            or not the input matrix is upper triangular.

        Returns
        -------
        BandedMatrix
            The new BandedMatrix.

        Notes
        -----
        Mostly stolen from sparse.coo_matrix.todia(), plus nan padding from
        BandedMatrix.from_ndarray() and nan triangles from
        BandedMatrix.roll_matrix().
        """
        # guess upper if None was passed
        if upper is None:
            upper = sparse.tril(spmatrix, k=-1).nnz == 0

        # check empty bins to be wiped later
        idx = np.where(
            np.squeeze(np.asarray(spmatrix.tocsr().getnnz(0) == 0)))[0]

        # logic for converting COO to DIA-style data
        n = spmatrix.shape[0]
        max_range = min(n, max_range)
        coo = spmatrix.tocoo()
        coo.sum_duplicates()
        ks = coo.col - coo.row
        keep_idx = np.abs(ks) <= max_range
        ks = ks[keep_idx]
        js = coo.col[keep_idx]
        vs = coo.data[keep_idx]
        data = np.zeros((max_range*2 + 1, n))
        data[ks + max_range, js] = vs

        # add padding and nan triangles to data
        data[np.tril(np.ones_like(data, dtype=bool), k=-max_range-1)] = np.nan
        data[np.triu(np.ones_like(data, dtype=bool), k=n-max_range)] = np.nan
        data = np.pad(data, ((1, 1), (0, 0)), mode='constant',
                      constant_values=np.nan)

        # construct BandedMatrix instance
        instance = cls((data, np.arange(-max_range - 1, max_range + 2)),
                       shape=spmatrix.shape)

        # wipe empty bins with nan
        instance[idx, :] = np.nan
        instance[:, idx] = np.nan

        # honor upper
        if upper:
            instance.make_upper()

        return instance

    @classmethod
    def is_bandedmatrix(cls, x):
        if type(x) == str:
            try:
                instance = sparse.load_npz(x)
            except (IOError, ValueError):
                return False
        else:
            instance = x
        if not isinstance(instance, sparse.dia_matrix):
            return False
        if not np.all(np.diff(instance.offsets) == 1):
            return False
        if not np.all(np.isnan(instance.data[0, :])):
            return False
        if not np.all(np.isnan(instance.data[-1, :])):
            return False
        return True

    def is_upper(self):
        return np.all(self.offsets >= -1)

    def max_range(self):
        return np.max(np.abs(self.offsets)) - 1

    @classmethod
    def load(cls, fname, **kwargs):
        if fname.endswith('.npy'):
            return cls(np.load(fname), **kwargs)
        return cls(sparse.load_npz(fname), **kwargs)

    def save(self, fname):
        check_outdir(fname)
        sparse.save_npz(fname, self)

    def copy(self):
        return self.from_dia_matrix(self)

    @classmethod
    def align(cls, *matrices):
        common_offsets = set.intersection(*(set(m.offsets) for m in matrices))
        res = []
        for m in matrices:
            extra_offsets = set(m.offsets) - common_offsets
            if extra_offsets:
                extra_offsets = np.array(list(extra_offsets))
                res.append(cls(
                    (np.delete(m.data, extra_offsets, axis=0),
                     np.delete(m.offsets, extra_offsets)), shape=m.shape))
            else:
                res.append(m)
        return res

    @staticmethod
    def roll_matrix(matrix, max_range=300):
        n = len(matrix)
        m = 2*max_range + 1
        row_idx, col_idx = np.ogrid[:n, :m]
        col_idx = (col_idx + np.arange(n)[:, np.newaxis] - max_range) % n
        rolled_matrix = matrix[row_idx, col_idx].T[::-1, :]
        rolled_matrix[np.tril(np.ones_like(rolled_matrix, dtype=bool),
                              k=-max_range-1)] = np.nan
        rolled_matrix[np.triu(np.ones_like(rolled_matrix, dtype=bool),
                              k=n-max_range)] = np.nan
        return rolled_matrix

    @staticmethod
    def make_mask(matrix, min_range=None, max_range=None, upper=False,
                  nan=False):
        """
        General-purpose function for creating masks for contact matrices.

        Parameters
        ----------
        matrix : np.ndarray
            Square contact matrix to make a mask for.
        min_range, max_range : int, optional
            Pass ints to specify a minimum and maximum allowed interaction
            range, in
            bin units.
        upper : bool
            Pass True to restrict the mask to the upper triangular entries.
        nan : bool
            Pass True to restrict the mask to non-NaN points.

        Returns
        -------
        np.ndarray
            The mask. Same shape as ``matrix``, with bool dtype.
        """
        mask = np.ones_like(matrix, dtype=bool)
        if min_range is not None:
            mask[np.triu(
                np.tril(np.ones_like(matrix, dtype=bool), k=min_range - 1),
                k=-(min_range - 1))] = False
        if max_range is not None:
            mask[np.triu(np.ones_like(matrix, dtype=bool),
                         k=max_range + 1)] = False
            mask[np.tril(np.ones_like(matrix, dtype=bool),
                         k=-(max_range + 1))] = False
        if upper:
            mask[np.tril(np.ones_like(matrix, dtype=bool), k=-1)] = False
        if nan:
            mask[np.isnan(matrix)] = False
        return mask

    @classmethod
    def apply(cls, f, *args, **kwargs):
        """
        Applies a function that takes in raw banded matrices (rectangular dense
        arrays) and returns raw banded matrices to one or more BandedMatrices.

        Parameters
        ----------
        f : function
            The function to apply. At least its first positional arg (or as many
            as all of them) should be an ``np.ndarray`` corresponding to the
            ``data`` array of a BandedMatrix. It should return one
            ``np.ndarray`` or a tuple of ``np.ndarray`` corresponding to the
            ``data`` array(s) of the resulting BandedMatrix(ces).
        *args : positional arguments
            Passed through to ``f()``. Instances of BandedMatrix will be passed
            as just their ``data`` arrays. If multiple BandedMatrices are
            passed, no attempt will be made to align them - a ValueError will be
            raised if they are not aligned.
        **kwargs : keyword arguments
            Passed through to ``f()``. No conversion from BandedMatrix will be
            attempted.

        Returns
        -------
        BandedMatrix or tuple of BandedMatrix
            If ``f()`` returns a single ``np.ndarray``, ``BandedMatrix.apply(f,
            arg1, ...)`` will return one BandedMatrix, whose shape and offsets
            will be taken from ``arg1`` (which must always be an BandedMatrix)
            and whose data will be taken from ``f(arg1.data, ...)``. If ``f()``
            returns a tuple of ``np.ndarray``, they will each be repackaged into
            an BandedMatrix in this fashion, and the tuple of new BandedMatrix
            will be returned.
        """
        if not all(np.array_equal(args[0].offsets, args[i].offsets)
                   for i in range(1, len(args)) if isinstance(args[i], cls)):
            raise ValueError('inputs not aligned')
        res_data = f(*(arg.data if isinstance(arg, cls) else arg
                       for arg in args), **kwargs)
        if isinstance(res_data, np.ndarray):
            return cls((res_data, args[0].offsets), shape=args[0].shape)
        return (cls((d, args[0].offsets), shape=args[0].shape)
                for d in res_data)

    @classmethod
    def max(cls, *matrices):
        """
        Returns the element-wise maximum of a list of BandedMatrices.

        Parameters
        ----------
        *matrices : sequence of BandedMatrix
            The matrices to compute the maximum of. Must already be aligned.

        Returns
        -------
        BandedMatrix
            The maximum.

        Notes
        -----
        If the matrices aren't aligned, consider:

            BandedMatrix.max(BandedMatrix.align(bm_a, bm_b, bm_c, ...))

        """
        return cls.apply(lambda *x: np.max(x, axis=0), *matrices)

    def __eq__(self, other):
        res = self.copy()
        if isinstance(other, sparse.dia_matrix):
            self.align(res, other)
            res.data = res.data == other.data
        else:
            res.data = res.data == other
        return res

    def __lt__(self, other):
        res = self.copy()
        if isinstance(other, sparse.dia_matrix):
            self.align(res, other)
            res.data = res.data < other.data
        else:
            res.data = res.data < other
        return res

    def __gt__(self, other):
        res = self.copy()
        if isinstance(other, sparse.dia_matrix):
            self.align(res, other)
            res.data = res.data > other.data
        else:
            res.data = res.data > other
        return res

    def __le__(self, other):
        res = self.copy()
        if isinstance(other, sparse.dia_matrix):
            self.align(res, other)
            res.data = res.data <= other.data
        else:
            res.data = res.data <= other
        return res

    def __ge__(self, other):
        res = self.copy()
        if isinstance(other, sparse.dia_matrix):
            self.align(res, other)
            res.data = res.data >= other.data
        else:
            res.data = res.data >= other
        return res

    def __and__(self, other):
        res = self.copy()
        if isinstance(other, sparse.dia_matrix):
            self.align(res, other)
            res.data = res.data & other.data
        else:
            raise NotImplementedError('type coercion not implemented yet')
        return res

    def __or__(self, other):
        res = self.copy()
        if isinstance(other, sparse.dia_matrix):
            self.align(res, other)
            res.data = res.data | other.data
        else:
            raise NotImplementedError('type coercion not implemented yet')
        return res

    def __xor__(self, other):
        res = self.copy()
        if isinstance(other, sparse.dia_matrix):
            self.align(res, other)
            res.data = res.data ^ other.data
        else:
            raise NotImplementedError('type coercion not implemented yet')
        return res

    def __invert__(self):
        res = self.copy()
        res.data = ~res.data
        return res

    def __add__(self, other):
        res = self.copy()
        if isinstance(other, sparse.dia_matrix):
            self.align(res, other)
        res.data += other
        return res

    def __sub__(self, other):
        res = self.copy()
        if isinstance(other, sparse.dia_matrix):
            self.align(res, other)
        res.data -= other
        return res

    def __mul__(self, other):
        res = self.copy()
        if isinstance(other, sparse.dia_matrix):
            self.align(res, other)
        res.data *= other
        return res

    def __div__(self, other):
        res = self.copy()
        if isinstance(other, sparse.dia_matrix):
            self.align(res, other)
        res.data /= other
        return res

    def __pow__(self, power, modulo=None):
        res = self.copy()
        res.data = res.data**power
        return res

    def log(self):
        res = self.copy()
        res.data = np.log(res.data)
        return res

    def flatten(self):
        return self.data.flatten()

    def ravel(self):
        return self.data.ravel()

    def where(self, a, b):
        cond = self.copy()
        cond.data = cond.data.astype(bool)
        res = self.copy()
        res.data = res.data.astype(np.promote_types(
            a.data.dtype if isinstance(a, sparse.dia_matrix) else type(a),
            b.data.dtype if isinstance(b, sparse.dia_matrix) else type(b)
        ))
        if isinstance(a, sparse.dia_matrix):
            self.align(res, a)
            res.data[cond.data] = a.data[cond.data]
        else:
            res.data[cond.data] = a
        if isinstance(b, sparse.dia_matrix):
            self.align(res, b)
            res.data[~cond.data] = b.data[~cond.data]
        else:
            res.data[~cond.data] = b
        return res

    def data_indices(self, key):
        # allow indexing with boolean BandedMatrix e.g., bm[bm > 10]
        if isinstance(key, BandedMatrix):
            if key.data.dtype != np.bool:
                raise IndexError('if indexing with BandedMatrix, dtype must '
                                 'be bool')
            return key.data

        # catch invalid key
        if not (type(key) in [tuple, list] and len(key) == 2):
            raise IndexError('must use tuple of length 2 to index into matrix')

        # correct trivial slices (to work with ogrid later)
        rowslice = slice(None, self.shape[0], None) \
            if type(key[0]) == slice and key[0] == slice(None, None, None) \
            else key[0]
        colslice = slice(None, self.shape[1], None) \
            if type(key[1]) == slice and key[1] == slice(None, None, None) \
            else key[1]

        if type(key[0]) == slice and type(key[1]) == slice:
            # if both items are slices then we just use ogrid on both
            row_idx, col_idx = np.ogrid[rowslice, colslice]
        elif type(key[1]) == slice:
            # column index is a slice
            col_idx = np.ogrid[colslice][np.newaxis, :]
            # set of rows
            if hasattr(key[0], '__len__'):
                row_idx = np.array([key[0]]).T
            # single row, presumably
            else:
                row_idx = key[0]
        elif type(key[0]) == slice:
            # row index is a slice
            row_idx = np.ogrid[rowslice][:, np.newaxis]
            # set of cols
            if hasattr(key[1], '__len__'):
                col_idx = np.array([key[1]])
            # single col, presumably
            else:
                col_idx = key[1]
        else:
            # arbitrary element (includes tuple of array)
            row_idx = np.asarray(key[0])
            col_idx = np.asarray(key[1])

        row_idx = (col_idx - row_idx + np.sum(self.offsets < 0))
        row_idx, col_idx = np.unravel_index(np.ravel_multi_index(
            (row_idx, col_idx), self.data.shape, mode='clip'), self.data.shape)
        return row_idx, col_idx

    def __getitem__(self, item):
        return self.data[self.data_indices(item)]

    def __setitem__(self, key, value):
        self.data[self.data_indices(key)] = value

    def reshape(self, shape, order='C'):
        raise NotImplementedError()

    def make_upper(self, pad=True):
        removed_diags = np.where(self.offsets < 0)[0]
        self.data = np.delete(self.data, removed_diags, axis=0)
        self.offsets = np.delete(self.offsets, removed_diags)
        if pad:
            self.data = np.pad(self.data, ((1, 0), (0, 0)), mode='constant',
                               constant_values=np.nan)
            self.offsets = np.pad(self.offsets, ((1, 0), ), mode='constant',
                                  constant_values=-1)

    def symmetrize(self):
        if np.sum(self.offsets < 0):
            self.make_upper(pad=False)
        assert self.offsets[0] == 0
        row_idx, col_idx = np.ogrid[:self.data.shape[0]-1, :self.data.shape[1]]
        row_idx = self.data.shape[0] - row_idx - 1
        col_idx = col_idx + row_idx
        row_idx, col_idx = np.unravel_index(np.ravel_multi_index(
            (row_idx, col_idx), self.data.shape, mode='wrap'), self.data.shape)
        self.data = np.concatenate([self.data[row_idx, col_idx], self.data])
        self.offsets = np.concatenate([-self.offsets[:0:-1], self.offsets])
        return self

    def deconvolute(self, bias, invert=False):
        """
        Applies a bias vector to both the rows and the columns of the
        BandedMatrix.

        Parameters
        ----------
        bias : np.ndarray
            Bias vector to apply. Should match size of BandedMatrix.
        invert : bool
            Pass True to invert the bias vector (divide the BandedMatrix by the
            the outer product of the bias vector with itself). Pass False to
            multiply the BandedMatrix by the outer product of the bias vector
            with itself.

        Returns
        -------
        BandedMatrix
            The result of the deconvolution.

        Examples
        --------
        >>> import numpy as np
        >>> from hic3defdr.util.banded_matrix import BandedMatrix
        >>> a = np.arange(16).reshape(4, 4).astype(float)
        >>> a += a.T
        >>> bias = np.sqrt(np.sum(a, axis=0))
        >>> b = ((a / bias).T / bias).T
        >>> bm = BandedMatrix.from_ndarray(b, max_range=3)
        >>> bm = bm.deconvolute(bias)
        >>> np.allclose(bm.toarray(), a)
        True
        """
        if invert:
            bias = 1 / bias
        csr = self.tocsr()
        bias_csr = sparse.diags([bias], [0])
        biased_csr = bias_csr.dot(csr)
        biased_csr = biased_csr.dot(bias_csr)
        return self.from_sparse(
            biased_csr, max_range=self.max_range(), upper=self.is_upper())



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from hic3defdr import HiC3DeFDR
from hic3defdr.lowess import lowess_fit


def main():
    # load data
    h = HiC3DeFDR.load('output-fast')
    ys = h.load_data('disp_per_dist')[:, 0]
    old_fn = h.load_disp_fn('ES')

    # weighted lowess, depends on xs, ys, and window size w
    n = len(ys)
    xs = np.arange(n)
    idx = np.isfinite(ys)
    w = 20

    # compute rolling var
    var = np.zeros_like(ys)
    var[idx] = pd.Series(ys[idx]).rolling(window=w, center=True).var()

    # convert to precision
    prec = 1 / var

    # scale to make smallest precision 1
    min_prec = np.nanmin(prec)
    scaled_prec = prec * (1 / min_prec)

    # fill left and right side nan's
    # get the first non-nan precision (for filling left side)
    max_prec = scaled_prec[np.argmax(np.isfinite(scaled_prec))]
    max_fill_idx = np.isnan(scaled_prec) & (xs < n/2)
    min_fill_idx = np.isnan(scaled_prec) & (xs > n/2)
    scaled_prec[max_fill_idx] = max_prec
    # fill right side with 1s
    scaled_prec[min_fill_idx] = 1

    # floor and convert to int
    floored_prec = np.floor(scaled_prec).astype(int)

    # print the sum of the weights - needs to be reasonably small
    print(np.nansum(floored_prec))

    # create duplicated data
    expanded_xs = []
    expanded_ys = []
    for i in range(n):
        m = floored_prec[i]
        if not np.isfinite(m):
            continue
        expanded_xs.extend([xs[i]] * m)
        expanded_ys.extend([ys[i]] * m)

    # lowess fit duplicated data
    fn = lowess_fit(np.array(expanded_xs), np.array(expanded_ys), frac=0.2)

    # plots
    ax = plt.gca()
    ax2 = ax.twinx()
    ax.scatter(xs[idx], ys[idx], label='raw dispersion', color='C0')
    ax.plot(xs[idx], fn(xs[idx]), label='smoothed dispersion', color='C1', linewidth=3)
    ax2.plot(xs[idx], floored_prec[idx], label='precision', color='C2')
    ax.set_xlabel('distance')
    ax.set_ylabel('dispersion')
    ax2.set_ylabel('precision')
    ax.legend()
    plt.savefig('var_test.png')

    ax.plot(xs[idx], old_fn(xs[idx]), label='old dispersion', color='C3', linewidth=3)
    ax.set_xlim((0, 35))
    ax.set_ylim((0, 0.01))
    ax.legend()
    plt.savefig('var_test_zoomed.png')


if __name__ == '__main__':
    main()



import numpy as np
import scipy.sparse as sparse


def kr_balance(array, tol=1e-6, x0=None, delta=0.1, ddelta=3, fl=1,
               max_iter=3000):
    """
    Balances a matrix via Knight-Ruiz matrix balancing, using sparse matrix
    operations.

    Parameters
    ----------
    array : np.ndarray or scipy.sparse.spmatrix
        The matrix to balance. Must be 2 dimensional and square. Will be
        symmetrized using its upper triangular entries.
    tol : float
        The error tolerance.
    x0 : np.ndarray, optional
        The initial guess for the bias vector. Pass None to use a vector of all
        1's.
    delta, ddelta : float
        How close/far the balancing vectors can get to/from the positive cone,
        in terms of a relative measure on the size of the elements.
    fl : int
        Determines whether or not internal convergence statistics will be
        printed to standard output.
    max_iter : int
        The maximum number of iterations to perform.

    Returns
    -------
    sparse.csr_matrix, np.ndarray, np.ndarray
        The CSR matrix is the balanced matrix. It will be upper triangular if
        ``array`` was upper triangular, otherwise it will by symmetric. The
        first np.ndarray is the bias vector. The second np.ndarray is a list of
        residuals after each iteration.

    Notes
    -----
    The core logic of this implementation is transcribed (by Dan Gillis) from
    the original Knight and Ruiz IMA J. Numer. Anal. 2013 paper and differs only
    slightly from the implementation in Juicer/Juicebox (see comments).

    It then uses the "sum factor" approach from the Juicer/Juicebox code to
    scale the bias vectors up to match the scale of the original matrix (see
    comments).

    Overall, this function is not nan-safe, but you may pass matrices that
    contain empty rows (the matrix will be shrunken before balancing, but all
    outputs will match the shape of the original matrix).

    This function does not perform any of the logic used in Juicer/Juicebox to
    filter out sparse rows if the balancing fails to converge.

    If the max_iter limit is reached, this function will return the current best
    balanced matrix and bias vector - no exception will be thrown. Callers can
    compare the length of the list of residuals to max_iter - if they are equal,
    the algorithm has not actually converged, and the caller may choose to
    perform further filtering on the matrix and try again. The caller is also
    free to decide if the matrix is "balanced enough" using any other criteria
    (e.g., variance of nonzero row sums).
    """
    # check to see if input is upper triangular
    triu = sparse.tril(array, k=-1).nnz == 0

    # convert to CSR and symmetrize
    array = sparse.triu(sparse.csr_matrix(array))
    array += array.transpose() - sparse.diags([array.diagonal()], [0])

    # shrink matrix, storing copy of original full matrix
    nonzero = array.getnnz(1) > 0
    full_array = array.copy()
    array = array[nonzero, :][:, nonzero]

    # this block was copied from lib5c.algorithms.knight_ruiz.kr_balance() which
    # was transcribed by Dan Gillis from the Matlab source provided in the
    # supplement of Knight and Ruiz IMA J. Numer. Anal. 2013
    # the only changes made in this block (relative to the lib5c version) are
    # that the dot products have been rewritten to work with sparse matrices
    # the Matlab/Python implementation is nearly equivalent to the Java
    # function compteKRNormVector() at
    # https://github.com/theaidenlab/Juicebox/blob/
    # 5a56089c63957cb15401ea7906ab77e242dfd755/src/juicebox/tools/utils/
    # original/NormalizationCalculations.java#L138
    # except that
    # 1) the Matlab/Python implementation has a concept of Delta/ddelta, which
    #    serves as an upper limit analogue to delta
    # 2) the Java version stops infinite iteration using a not_changing counter,
    #    which terminates the iteration when rho_km1 hasn't changed
    #    significantly in 100 iterations and returns a failing result; the
    #    Python version uses a max_iter counter, which simply sets a maximum
    #    number of iterations to perform but does not return a failing result if
    #    the maximum number of iterations is exceeded; the Matlab implementation
    #    does neither and will instead loop forever
    dims = array.shape
    if dims[0] != dims[1] or len(dims) > 2:
        raise Exception
    n = dims[0]
    e = np.ones((n, 1))
    res = np.array([])
    if x0 is None:
        x0 = e.copy()
    g = np.float_(0.9)
    eta_max = 0.1
    eta = eta_max
    stop_tol = tol * 0.5
    # x0 is not used again, so do not need to copy object
    x = x0  # n x 1 array
    rt = tol ** 2
    v = x0 * array.dot(x)  # n x 1 array
    rk = 1 - v  # n x 1 array
    rho_km1 = np.dot(np.transpose(rk), rk)[0, 0]  # scalar value
    # no need to copy next two objects - are scalar values
    rout = rho_km1
    rold = rout
    mvp = 0
    i = 0
    if fl == 1:
        print('it in. it res')
    while rout > rt:
        if max_iter is not None and i > max_iter:
            break
        i += 1
        k = 0
        y = np.ones((n, 1))
        innertol = max(eta ** 2 * rout, rt)
        rho_km2 = None
        while rho_km1 > innertol:
            k += 1
            if k == 1:
                z = rk / v
                p = z.copy()
                rho_km1 = np.dot(np.transpose(rk), z)
            else:
                beta = rho_km1 / rho_km2
                p = z + beta * p
            w = x * array.dot(x * p) + v * p
            alpha = rho_km1 / np.dot(np.transpose(p), w)
            ap = alpha * p
            ynew = y + ap
            if np.min(ynew) <= delta:
                if delta == 0:
                    break
                ind = np.where(ap < 0)
                gamma = np.min((delta - y[ind]) / ap[ind])
                y = y + gamma * ap
                break
            if np.max(ynew) >= ddelta:
                # see above - need to code
                # also check that (i.e. gamma or ap is scalar)
                ind = np.where(ynew > ddelta)
                gamma = min((ddelta - y[ind]) / ap[ind])
                y = y + gamma * ap
                break
            y = ynew
            rk = rk - alpha * w
            rho_km2 = rho_km1
            z = rk / v
            rho_km1 = np.dot(np.transpose(rk), z)
        x = x * y
        v = x * array.dot(x)
        rk = 1 - v
        rho_km1 = np.dot(np.transpose(rk), rk)[0, 0]  # scalar value
        rout = rho_km1
        mvp = mvp + k + 1
        rat = rout / rold
        rold = rout
        res_norm = np.sqrt(rout)
        eta_0 = eta
        eta = g * rat
        if g * eta_0 ** 2 > 0.1:
            eta = max(eta, g * eta_0 ** 2)
        eta = max(min(eta, eta_max), stop_tol / res_norm)
        if fl == 1:
            print('{} {} {:.3e}'.format(i, k, res_norm))
            res = np.append(res, res_norm)
    del array

    # expand bias vector
    bias = np.zeros_like(nonzero, dtype=float)
    bias[nonzero] = np.squeeze(x)

    # compute balanced matrix
    bias_csr = sparse.diags([bias], [0])
    balanced = bias_csr.dot(full_array)
    balanced = balanced.dot(bias_csr)

    # compute "sum factor" and recompute balanced matrix
    # this is equivalent to the Java function getSumFactor() at
    # https://github.com/theaidenlab/Juicebox/blob/
    # 5a56089c63957cb15401ea7906ab77e242dfd755/src/juicebox/tools/utils/
    # original/NormalizationCalculations.java#L333
    sum_factor = np.sqrt(full_array.sum() / balanced.sum())
    bias *= sum_factor
    bias_csr = sparse.diags([bias], [0])
    balanced = bias_csr.dot(full_array)
    balanced = balanced.dot(bias_csr)
    del full_array

    # invert bias vector at nonzero positions
    bias[bias != 0] = 1 / bias[bias != 0]

    # if input was upper triangular, return upper triangular balanced matrix
    if triu:
        balanced = sparse.triu(balanced)
        balanced = balanced.tocsr()

    return balanced, bias, res



"""
Module for managing version naming scheme and supplying version information.

We use setuptools_scm to get version information from git. The version naming
scheme is specified by defining functions ``version_scheme()`` and
``local_scheme()`` in this module. We have written these so that they mimic
versioneer's "pep440" style. These functions are used in this module, and they
are also imported in ``setup.py`` at install time.

This module also provides the function ``get_version()``, which gets the version
if either setuptools_scm or our package is installed, and returns "unknown"
otherwise. Getting the version from setuptools_scm is primarily useful for
Docker image building (where it doesn't make sense to make the host install our
package just to obtain the version information to pass to the image build
process) and for editable installs (where having setuptools_scm installed is the
only way to get accurate version information).

When our package is installed, its version (the result of ``get_version()``) can
be accessed as ``<our_package>.__version__``.

This module can be run as a script to print version information to stdout.
"""


def version_scheme(version):
    return str(version.tag)


def local_scheme(version):
    if version.distance is None:
        return ''
    return '+%s.%s%s' % (version.distance, version.node,
                         '.dirty' if version.dirty else '')


def get_version():
    try:
        # our first choice would be to get the version from setuptools_scm if it
        # is installed (only way that works with editable installs)
        from setuptools_scm import get_version as scm_get_version
        version = scm_get_version(
            root='..',
            relative_to=__file__,
            version_scheme=version_scheme,
            local_scheme=local_scheme
        )
    except ImportError:
        try:
            try:
                # this works in Python 3.8
                from importlib.metadata import PackageNotFoundError, \
                    version as importlib_version
            except ImportError:
                # this works in Python 2 if our package is installed, since
                # it depends on importlib_metadata
                from importlib_metadata import PackageNotFoundError, \
                    version as importlib_version
            try:
                # we land here if either importlib.metadata or
                # importlib_metadata is available
                # we expect that this module is a submodule of our actual
                # package
                # we expect that __name__ is <our_package>._version
                # therefore, splitting on . should allow us to get the name of
                # our package
                version = importlib_version(__name__.split('.')[0])
            except PackageNotFoundError:
                # we will land here if our package isn't actually installed
                version = 'unknown'
        except ImportError:
            # we land here if neither importlib.metadata nor importlib_metadata
            # are available
            version = 'unknown'
    return version


def main():
    print(get_version())


if __name__ == '__main__':
    main()



import os
from six.moves.urllib.request import urlretrieve
from zipfile import ZipFile

from hic3defdr.util.printing import eprint


DEMO_DATA_URL = 'https://www.dropbox.com/sh/mq0fpnp4jz59wpo/AAD2FW1Tp_mVKCkxlJoZvxC8a?dl=1'  # noqa: E501
CHROMS = ['chr18', 'chr19']
CONDS = ['ES', 'NPC']
COND_EXT = ['clusters.json']
REPS = ['ES_1', 'ES_3', 'NPC_1', 'NPC_2']
REP_EXT = ['kr.bias', 'raw.npz']


def check_demo_data(dest_dir='~/hic3defdr-demo-data'):
    """
    Checks to see if all demo files are present.

    Parameters
    ----------
    dest_dir : str
        Path to destination directory to look for demo data files in.

    Returns
    -------
    bool
        True if all demo files are present, False otherwise.
    """
    dest_dir = os.path.expanduser(dest_dir)
    for chrom in CHROMS:
        for cond in CONDS:
            for ext in COND_EXT:
                path = os.path.join(
                    dest_dir, 'clusters', '%s_%s_%s' % (cond, chrom, ext))
                if not os.path.exists(path):
                    eprint('failed to find expected demo file at %s' % path)
                    return False
        for rep in REPS:
            for ext in REP_EXT:
                path = os.path.join(dest_dir, rep, '%s_%s' % (chrom, ext))
                if not os.path.exists(path):
                    eprint('failed to find expected demo file at %s' % path)
                    return False
    return True


def ensure_demo_data(dest_dir='~/hic3defdr-demo-data'):
    """
    Checks that all the demo data files are present to dest_dir, and downloads
    them if any are missing.

    Parameters
    ----------
    dest_dir : str
        Path to destination directory to download files to.
    """
    dest_dir = os.path.expanduser(dest_dir)
    if check_demo_data(dest_dir=dest_dir):
        eprint('demo data is present on disk, not re-downloading')
        return
    eprint('downloading demo data to temp.zip')
    urlretrieve(DEMO_DATA_URL, 'temp.zip')
    eprint('extracting demo data to %s' % dest_dir)
    try:
        with ZipFile('temp.zip', 'r') as z:
            z.extractall(dest_dir)
    finally:
        eprint('deleting temp.zip')
        os.remove('temp.zip')


def main():
    ensure_demo_data()


if __name__ == '__main__':
    main()



import numpy as np
import scipy.sparse as sparse

from hic3defdr.util.banded_matrix import BandedMatrix


def filter_sparse_rows_count(matrix, min_nnz=25, k=300):
    """
    Wipes the sparse bins (both rows and columns) of a matrix, where "sparse"
    means that the bin has less than ``min_nnz`` nonzero interactions with both
    the ``k`` nearest upstream and downstream bins.

    This function wipes with zeros and will eliminate the wiped positions if
    ``matrix`` is a CSR format sparse matrix.

    Parameters
    ----------
    matrix : np.ndarray, scipy.sparse.csr_matrix, or BandedMatrix
        The matrix to wipe sparse bins from.
    min_nnz : int
        The minimum number of nonzero contacts a bin has to make in either
        direction (upstream or downstream) to escape being wiped.
    k : int
        How many bins upstream or downstream to look when counting the number
        of nonzero contacts a bin makes in a given direction.

    Returns
    -------
    np.ndarray, scipy.sparse.csr_matrix, or BandedMatrix
        The filtered matrix.
    """
    matrix = matrix.copy()
    if min_nnz == 0 or k == 0:
        return matrix
    if isinstance(matrix, np.ndarray):
        # matrix already supports row/col indexing
        # BandedMatrix can be created directly from matrix
        bm = BandedMatrix.from_ndarray(matrix, max_range=k)
    else:
        if BandedMatrix.is_bandedmatrix(matrix):
            bm = matrix
        elif isinstance(matrix, sparse.spmatrix):
            if not isinstance(matrix, sparse.csr_matrix):
                # incoming matrix doesn't support row/col indexing, convert it
                matrix = matrix.tocsr()
            # create BandedMatrix directly from sparse matrix
            bm = BandedMatrix(matrix, max_range=k)
        else:
            raise ValueError('invalid input type')
    bm.symmetrize()
    us_rows = np.where((bm.offsets >= -k) & (bm.offsets < 0))[0]
    ds_rows = np.where((bm.offsets <= k) & (bm.offsets > 0))[0]
    deleted_indices = (np.sum(bm.data[us_rows, :] > 0, axis=0) < min_nnz) & \
                      (np.sum(bm.data[ds_rows, :] > 0, axis=0) < min_nnz)
    if isinstance(matrix, sparse.csr_matrix):
        keep_indices_csr = sparse.diags([~deleted_indices], [0], dtype=int)
        matrix = keep_indices_csr.dot(matrix)
        matrix = matrix.dot(keep_indices_csr)
    else:
        deleted_indices = np.where(deleted_indices)[0]
        matrix[:, deleted_indices] = 0
        matrix[deleted_indices, :] = 0
    return matrix
